<?php
namespace NjFbCloud\Src\Classes;


class DropboxApi {

    private static $option_key = 'dropbox';

    private static $api_ver = '2';

    /**
     * Gets authurl in order to get access token
     * Docs: https://developers.dropbox.com/oauth-guide
     * 
     * @return string
     */
    public static function authUrl()
    {
        return add_query_arg( array(
            'response_type' => 'code',
            'redirect_uri' => urlencode( njfb_cloud_dropbox_redirecturis() ),
            'client_id' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_id', '' ),
            'token_access_type' => 'offline',
            'state' => ''
        ), 'https://www.dropbox.com/oauth2/authorize' );
    }

    public static function requestToken( $code )
    {
        $res = wp_remote_post( 'https://api.dropbox.com/oauth2/token', array(
            'body'    => array(
                'code' => $code,
                'client_id' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_id', '' ),
                'client_secret' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_secret', '' ),
                'redirect_uri' => njfb_cloud_dropbox_redirecturis(),
                'grant_type' => 'authorization_code'
            ),
            'headers' => array(
                'Content-Type' => 'application/x-www-form-urlencoded',
            ),
            'timeout'     => 45,
        ) );
        if( ! is_wp_error( $res ) ) {
            $body = json_decode( wp_remote_retrieve_body( $res ), true );
            if( is_array( $body ) && isset( $body['access_token'] ) ) {
                $body = self::appendMoreTokenFields( $body );
                return $body;
            }
            return null;
        }
        return null;
    }
    public static function revoke( $access_token ) {
        $res = wp_remote_post( self::endpoint() . '/auth/token/revoke', array(
            'headers' => array(
                'Authorization' => 'Bearer ' . $access_token,
            ),
            'timeout' => 45,
        ) );
        return $res;
    }
    public static function createFolder( $bearer_token, $name, $parent = null ) {
        $body = array( 'path' => '/' . ltrim( $name, '/' ) );
        
        $headers = array(
            'Authorization' => 'Bearer ' . $bearer_token,
            'Content-Type' => 'application/json'
        );
        $res = wp_remote_post( self::endpoint() . '/files/create_folder_v2', array( 
            'body' => wp_json_encode( $body ),
            'headers' => $headers
        ) );
        $apiBody = json_decode( wp_remote_retrieve_body( $res ), true);
        return $apiBody;
    }
    private static function endpoint() {
        return 'https://api.dropboxapi.com/' . self::$api_ver;
    }
    private static function appendMoreTokenFields( $token )
    {
        $token['created_at'] = time();
        return $token;
    }
    public static function isExpired( $token )
    {
        return ((int)$token['created_at'] + (int)$token['expires_in'] - 20) < time();
    }
    public static function refreshToken( $refresh_token )
    {
        $res = wp_remote_post( 'https://api.dropboxapi.com/oauth2/token', array(
            'body' => array(
                'client_id' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_id', '' ),
                'client_secret' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_secret', '' ),
                'refresh_token' => $refresh_token,
                'grant_type' => 'refresh_token'
            ),
            'headers' => array(
                'Content-Type' => 'application/x-www-form-urlencoded',
            ),
        ));

        if( ! is_wp_error( $res ) ) {
            $body = json_decode( wp_remote_retrieve_body( $res ), true );
            if( is_array( $body ) && isset( $body['access_token'] ) ) {
                $body = self::appendMoreTokenFields( $body );
                $body['refresh_token'] = $refresh_token;
                return $body;
            } else {
                throw new \Exception($body['error'] . ' - ' . $body['error_description'] . '. Please disconnect then reconnect to try again.');
            }
            return null;
        }
        return null;
    }
    public static function isFolderAlive( $access_token, $path )
    {
        $url = 'https://api.dropboxapi.com/'.self::$api_ver.'/files/get_metadata';
        $headers = array(
            'Authorization' => 'Bearer ' . $access_token,
            'Content-Type' => 'application/json',
        );
        $body = json_encode(array('path' => $path));

        $response = wp_remote_post($url, array(
            'headers' => $headers,
            'body' => $body,
        ));

        if (is_wp_error($response)) {
            throw new \Exception( 'Error When Check Folder: ' . $response->get_error_message() );
        }

        $status_code = wp_remote_retrieve_response_code($response);
        if ($status_code === 409) {
            // Folder does not exist or is deleted
            return false;
        } elseif ($status_code !== 200) {
            throw new \Exception( 'Error When Check Folder, Some other error occurred: ' . wp_remote_retrieve_body($response) );
        }

        $body = wp_remote_retrieve_body($response);
        $folder_metadata = json_decode($body, true);

        // Check if the returned metadata indicates a deleted folder
        if (isset($folder_metadata['.tag']) && $folder_metadata['.tag'] === 'deleted') {
            return false;
        }

        return true;
    }
    public static function getFiles( $access_token, $path = '' ) {
        $url = self::endpoint() . '/files/list_folder';
        $headers = array(
            'Authorization' => 'Bearer ' . $access_token,
            'Content-Type' => 'application/json',
        );
        $body = json_encode(array('path' => $path));
    
        $response = wp_remote_post($url, array(
            'headers' => $headers,
            'body' => $body,
        ));
    
        if (is_wp_error($response)) {
            return 'Error: ' . $response->get_error_message();
        }
    
        $status_code = wp_remote_retrieve_response_code($response);
        if ($status_code !== 200) {
            return 'Error: ' . wp_remote_retrieve_body($response);
        }
    
        $body = wp_remote_retrieve_body($response);
        $folder_content = json_decode($body, true);
    
        return $folder_content;
    }
    public static function getFilesContinue( $access_token, $cursor) {
        $url = self::endpoint() . '/files/list_folder/continue';
        $headers = array(
            'Authorization' => 'Bearer ' . $access_token,
            'Content-Type' => 'application/json',
        );
        $body = json_encode(array('cursor' => $cursor));
    
        $response = wp_remote_post($url, array(
            'headers' => $headers,
            'body' => $body,
        ));
    
        if (is_wp_error($response)) {
            return 'Error: ' . $response->get_error_message();
        }
    
        $status_code = wp_remote_retrieve_response_code($response);
        if ($status_code !== 200) {
            return 'Error: ' . wp_remote_retrieve_body($response);
        }
    
        $body = wp_remote_retrieve_body($response);
        $folder_content = json_decode($body, true);
    
        return $folder_content;
    }
    public static function getFileContent( $access_token, $path ) {
        $url = 'https://content.dropboxapi.com/'.self::$api_ver.'/files/download';
        $headers = array(
            'Authorization' => 'Bearer ' . $access_token,
            'Dropbox-API-Arg' => json_encode(array('path' => $path)),
            'Content-Type' => 'text/plain'
        );
    
        $response = wp_remote_get($url, array(
            'timeout' => 120,
            'headers' => $headers,
        ));
    
        if (is_wp_error($response)) {
            throw new \Exception( 'Cant download file. Error #1 ' . $response->get_error_message() );
        }
    
        $status_code = wp_remote_retrieve_response_code($response);
        if ($status_code !== 200) {
            throw new \Exception( 'Cant download file. Error #2 ' . wp_remote_retrieve_body($response) );
        }
    
        $body = wp_remote_retrieve_body($response);
    
        return $body;
    }
}
