<?php
namespace NjFbCloud\Src\Classes;
class GoogleApiUploadResumable
{
    private $path;
    private $token;
    private $parent = null;

    private $session_uri;
    private $filesize;

    private $position = 0;
    private $length = 4194304; //4MB

    public function __construct( $args )
    {
        $args = wp_parse_args( $args, array(
            'path' => '',
            'token' => '',
            'parent' => null
        ) );
        $this->path = $args['path'];
        $this->token = $args['token'];
        $this->filesize = filesize( $this->path );
        
        $this->parent = $args['parent'];
    }

    public function setSessionUri() {
        $cache_key = 'njfb_cloud_session_uri_' . md5_file( $this->path );
        if( ($this->session_uri = get_site_transient( $cache_key )) === false  ) {
            $metadata = array(
                "name" => basename( $this->path )
            );
            if( ! is_null( $this->parent ) ) {
                $metadata['parents'] = array( $this->parent );
            }
            $res = wp_remote_post( 'https://www.googleapis.com/upload/drive/v3/files?uploadType=resumable', array(
                'headers' => array(
                    'Authorization' => 'Bearer '. $this->token,
                    'X-Upload-Content-Type' => mime_content_type( $this->path ),
                    'X-Upload-Content-Length' => $this->filesize,
                    'Content-Type' => 'application/json; charset=UTF-8',
                    // 'Content-Length' => ''
                ),
                'body' => wp_json_encode( $metadata )
            ) );
            $resHeaders = wp_remote_retrieve_headers( $res );
            $this->session_uri = isset( $resHeaders['location'] ) ? $resHeaders['location'] : null;
            if( ! is_null( $this->session_uri ) ) {
                set_site_transient( $cache_key, $this->session_uri, (time() + ( 7 * 24 * 3600 ) - 3600 ) ); // 1 week - 1 hour
            }
        }
    }
    public function clearSessionUri() {
        delete_site_transient( 'njfb_cloud_session_uri_' . md5_file( $this->path ) );
        $this->session_uri = false;
    }
    public function setPosition( $position ) {
        $this->position = $position;
    }
    public function setLength( $length ) {
        $this->length = $length;
    }
    
    public function uploadChunk() {
        $this->setSessionUri();
        $position = $this->position;
        $length = $this->length;

        $handler = fopen( $this->path, 'rb' );
        if( ! feof($handler) && $length > 0 ) {
            if( $position > 0 ) {
                fseek( $handler, $position );
            }
            $chunk = fread($handler, $length);
            $res = wp_remote_request( $this->session_uri, array(
                'method' => 'PUT',
                'headers' => array(
                    'Content-Length' => $length,
                    'Content-Range' => "bytes {$position}-". ($position + $length - 1) ."/{$this->filesize}",
                ),
                'body' => $chunk
            ) );
            $statusCode = wp_remote_retrieve_response_code( $res );
            // $resHeaders = wp_remote_retrieve_headers( $res );
            $resBody = wp_remote_retrieve_body( $res );
            $resBody = json_decode( $resBody, true );
            if ($statusCode == 308) {
                $header_range = wp_remote_retrieve_header( $res, 'range' );
                preg_match( '#bytes=([0-9]+)-([0-9]+)#', $header_range, $m );
                $range_from = $m[1];
                $range_to = $m[2];

                $position = $range_to + 1;

                $r = $this->filesize - $position;
                $length = ($r >= $length) ? $length : $r;

                return [
                    'position' => $position,
                    'length' => $length,
                    'finish' => false,
                    'statusCode' => $statusCode
                ];
            } elseif( $statusCode == 403 ) {
                // Rate Limit Exceeded
                $this->clearSessionUri();
                return [
                    'finish' => false,
                    'statusCode' => $statusCode,
                    'mess' => 'Rate Limit Exceeded'
                ];
            } elseif( $statusCode == 404 ) {
                //the upload session has expired, the upload will be restarted from the beginning.
                $this->clearSessionUri();
                return [
                    'finish' => false,
                    'statusCode' => $statusCode,
                    'mess' => 'the upload session has expired, the upload will be restarted from the beginning.'
                ];
            } elseif( $statusCode == 200 || $statusCode == 201 ) {
                // Done
                $this->clearSessionUri();
                return [
                    'finish' => true,
                    'statusCode' => $statusCode,
                    'id' => $resBody['id']
                ];
            } elseif( $statusCode >= 400 && $statusCode <= 499 ) {
                // restart the upload
                $this->clearSessionUri();
                return [
                    'finish' => false,
                    'statusCode' => $statusCode,
                    'mess' => 'Restart the upload'
                ];
            } else {
                // retry
                return [
                    'position' => $position,
                    'length' => $length,
                    'finish' => false,
                    'statusCode' => $statusCode
                ];
            }
        }
        fclose( $handler );
    }
}
