<?php
namespace NjFbCloud\Src\Classes;

use NjFbCloud\Src\Enum\GoogleDriveMime;

/**
 * Class GoogleDriveApi
 * Provides methods for interacting with Google Drive API.
 */
class GoogleDriveApi {

    /**
     * Key for accessing Google Drive settings
     * @var string
     */
    private static $option_key = 'google_drive';

    /**
     * Google API version
     * @var string
     */
    private static $api_ver = 'v3';

    /**
     * Requests token from Google API using authorization code
     * 
     * @param string $code The authorization code
     * @return array|null The token if successful, null otherwise
     */
    public static function requestToken( $code )
    {
        $redirect_uri = njfb_cloud_ggdrive_redirecturis();
        // Making a POST request to Google API token endpoint
        $res = wp_remote_post( 'https://www.googleapis.com/oauth2/'. self::$api_ver .'/token', array(
            'body'    => array(
                'code' => $code,
                'client_id' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_id', '' ),
                'client_secret' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_secret', '' ),
                'redirect_uri' => $redirect_uri,
                'grant_type' => 'authorization_code'
            ),
            'headers' => array(
                'Content-Type' => 'application/x-www-form-urlencoded',
            ),
            'timeout'     => 45,
        ) );

        if( ! is_wp_error( $res ) ) {
            $body = json_decode( wp_remote_retrieve_body( $res ), true );
            if( is_array( $body ) && isset( $body['access_token'] ) ) {
                $body = self::appendMoreTokenFields( $body );
                return $body;
            }
            return null;
        }
        return null;
    }

    /**
     * Revokes access token from Google API
     * 
     * @param string $token The access token
     * @return bool True if successful, false otherwise
     */
    public static function revokeAccess( $token )
    {
        $res = wp_remote_post( 'https://accounts.google.com/o/oauth2/revoke', array(
            'body'    => array(
                'token' => $token,
            ),
            'headers' => array(
                'Content-Type' => 'application/x-www-form-urlencoded',
            ),
        ) );
        if( ! is_wp_error( $res ) )
        {
            return true;
        }
        return false;
    }

    /**
     * Refreshes access token using refresh token
     * 
     * @param string $refresh_token The refresh token
     * @return array|null The refreshed token if successful, null otherwise
     */
    public static function refreshToken( $refresh_token )
    {
        $res = wp_remote_post( 'https://accounts.google.com/o/oauth2/token', array(
            'body' => array(
                'client_id' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_id', '' ),
                'client_secret' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_secret', '' ),
                'refresh_token' => $refresh_token,
                'grant_type' => 'refresh_token'
            ),
            'headers' => array(
                'Content-Type' => 'application/x-www-form-urlencoded',
            ),
        ));

        if( ! is_wp_error( $res ) ) {
            $body = json_decode( wp_remote_retrieve_body( $res ), true );
            if( is_array( $body ) && isset( $body['access_token'] ) ) {
                $body = self::appendMoreTokenFields( $body );
                $body['refresh_token'] = $refresh_token;
                return $body;
            } else {
                throw new \Exception($body['error'] . ' - ' . $body['error_description'] . '. Please disconnect then reconnect to try again.');
            }
            return null;
        }
        return null;
    }

    /**
     * Generates authorization URL for Google API
     * 
     * @return string The authorization URL
     */
    public static function authUrl()
    {
        $scope = 'https://www.googleapis.com/auth/drive';
        $redirect_uri = njfb_cloud_ggdrive_redirecturis();

        return add_query_arg( array(
            'response_type' => 'code',
            'redirect_uri' => urlencode( $redirect_uri ),
            'client_id' => njfb_cloud_get_settings_by_key( self::$option_key, 'client_id', '' ),
            'scope' => $scope,
            'access_type' => 'offline',
            // 'approval_prompt' => 'consent',
            // 'prompt' => 'force',
            'state' => ''
        ), 'https://accounts.google.com/o/oauth2/auth' );
    }

    // Checks if token is expired
    public static function isExpired( $token )
    {
        return ((int)$token['created_at'] + (int)$token['expires_in'] - 20) < time();
    }

    // Gets Google Drive API metadata URI
    public static function getApiMetaUri()
    {
        return sprintf( 'https://www.googleapis.com/drive/%s/', self::$api_ver );
    }

    // Gets Google Drive API upload URI
    public static function getApiUploadUri()
    {
        return sprintf( 'https://www.googleapis.com/upload/drive/%s/files', self::$api_ver );
    }

    // Gets file media content from Google Drive
    public static function getFileMediaContent( $bearer_token, $file_id )
    {
        $url = add_query_arg( array( 'alt' => 'media' ), self::getApiMetaUri() . 'files/' . $file_id );
        $res = wp_remote_get( $url, array(
            'headers'     => array(
                'Authorization' => 'Bearer ' . $bearer_token,
            ),
        ) );
        if( ! is_wp_error( $res ) ) {
            $body = wp_remote_retrieve_body( $res );
            return $body;
        }
        return false;
    }
    // Adds permission for file access
    public static function addFilePermisson( $bearer_token, $file_id ) {
        $url = sprintf( 'https://www.googleapis.com/drive/%s/files/%s/permissions', self::$api_ver, $file_id );
        $res = wp_remote_post( $url, array(
            'headers'     => array(
                'Authorization' => 'Bearer ' . $bearer_token,
                'Content-Type' => 'application/json'
            ),
            'body' => wp_json_encode( array(
                'type' => 'anyone',
                'value' => 'anyone',
                'role' => 'reader'
            ) )
        ) );
        if( ! is_wp_error( $res ) ) {
            $body = json_decode( wp_remote_retrieve_body( $res ), true );
            return $body;
        }
        return false;
    }

    // Gets file metadata from Google Drive
    public static function getFileMeta( $bearer_token, $file_id )
    {
        $url = self::getApiMetaUri() . 'files/' . $file_id;
        $res = wp_remote_get( $url, array(
            'headers'     => array(
                'Authorization' => 'Bearer ' . $bearer_token,
            ),
        ) );
        if( ! is_wp_error( $res ) ) {
            $body = json_decode( wp_remote_retrieve_body( $res ), true );
            return $body;
        }
        return false;
    }
    public static function getFiles( $bearer_token, $mimes = GoogleDriveMime::Folder, $fields = array(), $parent = 'root', $page_token = null, $page_size = 100 )
    {
        $page_size = (int)$page_size;

        $return = array(
            'files' => array(),
            'page_token' => null
        );
        //mimes
        $mime_q = '';
        if( is_array( $mimes ) ) {
            $mime_q = array();
            foreach( $mimes as $m ) {
                $mime_q[] = 'mimeType=\''.$m.'\'';
            }
            $mime_q = '('.implode( ' or ', $mime_q ).')';
        } else {
            $mime_q = 'mimeType=\''.$mimes.'\'';            
        }
        //fields
        //https://developers.google.com/drive/api/reference/rest/v3/files#File
        if( ! is_array( $fields ) && $fields !== '*' ) {
            $fields = array($fields);
        }
        if( is_array( $fields ) ) {
            $fields = wp_parse_args( $fields, array('id', 'name', 'parents', 'createdTime', 'modifiedTime') );
            $field_q = implode(', ', $fields);
        } else {
            $field_q = $fields;
        }

        $args = array(
            // 'q' => '\''.$parent.'\' in parents and trashed=false and mimeType=\''.$mimes.'\'',
            'q' => '\''.$parent.'\' in parents and trashed=false' . ((!empty($mime_q)) ? ' and ' . $mime_q : ''),
            'orderBy' => 'name_natural',
            'fields' => 'nextPageToken, files('.$field_q.')',
            'pageSize' => $page_size
        );
        if( ! is_null( $page_token ) ) {
            $args['pageToken'] = $page_token;
        }
        $endpoint = add_query_arg( 
            $args,
            self::getApiMetaUri() . 'files'
        );
        $res = wp_remote_get( $endpoint, array(
            'headers'     => array(
                'Authorization' => 'Bearer ' . $bearer_token,
            ),
        ) );
        
        if( ! is_wp_error( $res ) ) {
            $body = json_decode( wp_remote_retrieve_body( $res ), true );
            if( isset( $body['files'] ) ) {
                $return['files'] = $body['files'];
                if( isset( $body['nextPageToken'] ) ) {
                    $return['page_token'] = $body['nextPageToken'];
                }
            }
        } else {
            // print_r($res);
            // exit;
        }
        return $return;
    }
    public static function createFolder( $bearer_token, $name, $parent = null ) {
        $body = array( 'name' => $name, 'mimeType' => GoogleDriveMime::Folder );
        if( ! is_null( $parent ) ) {
            $body['parents'] = array( $parent );
        }

        $headers = array(
            'Authorization' => 'Bearer ' . $bearer_token,
            'Content-Type' => 'application/json'
        );
        $res = wp_remote_post( self::getApiMetaUri() . 'files', array( 
            'body' => wp_json_encode( $body ),
            'headers' => $headers
        ) );
        $apiBody = json_decode( wp_remote_retrieve_body( $res ), true);
        return $apiBody;
    }
    public static function renameFolder( $bearer_token, $id, $new_name) {
        $headers = array(
            'Authorization' => 'Bearer ' . $bearer_token,
            'Content-Type' => 'application/json'
        );
        $body = array( 'name' => $new_name );
        $res = wp_remote_request( self::getApiMetaUri() . 'files/' . $id, array( 
            'body' => wp_json_encode( $body ),
            'headers' => $headers,
            'method' => 'PATCH'
        ) );
        $apiBody = json_decode( wp_remote_retrieve_body( $res ), true);
        return $apiBody;
    }
    public static function updateParentFolder( $bearer_token, $id, $new_parent) {
        $headers = array(
            'Authorization' => 'Bearer ' . $bearer_token,
            'Content-Type' => 'application/json'
        );
        $url = add_query_arg( array( 'addParents' => $new_parent ), self::getApiMetaUri() . 'files/' . $id );
        $res = wp_remote_request( $url, array( 
            'headers' => $headers,
            'method' => 'PATCH'
        ) );
        $apiBody = json_decode( wp_remote_retrieve_body( $res ), true);
        return $apiBody;
    }
    public static function deleteFolder( $bearer_token, $id ) {
        $headers = array(
            'Authorization' => 'Bearer ' . $bearer_token,
            'Content-Type' => 'application/json'
        );
        $res = wp_remote_request( self::getApiMetaUri() . 'files/' . $id, array( 
            'headers' => $headers,
            'method'     => 'DELETE'
        ) );
        $apiBody = wp_remote_retrieve_body($res);
        return $apiBody;
    }
    public static function uploadTypeMedia( $bearer_token, $file_path ) {
        $file_content = file_get_contents( $file_path );
        $mime_type = mime_content_type($file_path);

        $apiURL = add_query_arg( array( 'uploadType' => 'media' ), self::getApiUploadUri() );
        $headers = array(
            'Content-Type' => $mime_type,
            'Authorization' => 'Bearer '. $bearer_token
        );
        $res = wp_remote_post( $apiURL, array(
            'headers' => $headers,
            'body' => $file_content
        ) );
        return json_decode( wp_remote_retrieve_body($res), true );
    }
    public static function uploadTypeMultipart( $bearer_token, $file_path, $to_folder_id = null ) {
        $file_content = file_get_contents( $file_path );
        $mime_type = mime_content_type($file_path);

        $apiURL = add_query_arg( array( 'uploadType' => 'multipart' ), self::getApiUploadUri() );

        $metadata = array(
            "name" => basename( $file_path ),
            "mimeType" => $mime_type
        );
        if( ! is_null( $to_folder_id ) ) {
            $metadata["parents"] = array( $to_folder_id );
        }

        $boundary = "xxxxxxxxxx";
        $data = "--" . $boundary . "\r\n";
        $data .= "Content-Type: application/json; charset=UTF-8\r\n\r\n";
        $data .= json_encode( $metadata) . "\r\n";
        $data .= "--" . $boundary . "\r\n";
        $data .= "Content-Transfer-Encoding: base64\r\n\r\n";
        $data .= base64_encode($file_content);
        $data .= "\r\n--" . $boundary . "--";

        $headers = array(
            'Content-Type' => 'multipart/related; boundary=' . $boundary,
            'Authorization' => 'Bearer '. $bearer_token
        );
        $res = wp_remote_post( $apiURL, array(
            'headers' => $headers,
            'body' => $data
        ) );
        return json_decode( wp_remote_retrieve_body($res), true );
    }
    public static function isFolderAlive( $bearer_token, $folder_id ) {
        $url = "https://www.googleapis.com/drive/v3/files/$folder_id";
    
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_HTTPGET, 1);
        curl_setopt($ch, CURLOPT_HTTPHEADER, array(
            "Authorization: Bearer $bearer_token"
        ));
        $response = curl_exec($ch);
        $http_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);
    
        if ($http_code == 404) {
            // Folder is deleted
            return false;
        } elseif ($http_code == 200) {
            // Folder exists
            return true;
        } else {
            // Handle other HTTP codes or errors
            // Log or return an error message as needed
            return null;
        }
    }

    /**
     * Subscribe to Google Drive changes with a webhook using wp_remote_post.
     *
     * @param string $bearerToken The OAuth 2.0 Bearer Token.
     * @param string $webhookUrl The URL of the webhook endpoint.
     * @return mixed The response from the Google Drive API or WP_Error on failure.
     */
    public static function subscribeToDriveChanges($bearerToken, $webhookUrl) {
        // Get the start page token
        $startPageToken = self::getStartPageToken($bearerToken);
        if (!$startPageToken) {
            return new \WP_Error( 'cant_get_start_page_token', 'Failed to retrieve start page token' );
        }
        
        // Google Drive API endpoint for creating a notification channel
        $url = sprintf('https://www.googleapis.com/drive/%s/changes/watch', self::$api_ver);
        $url = add_query_arg( array( 'pageToken' => $startPageToken ), $url );
        $channelId = self::generateUUID();

        // Create the data array for the POST request
        $data = [
            'id' => $channelId,
            'type' => 'web_hook',
            'address' => $webhookUrl
        ];

        // Encode the data array to JSON format
        $dataString = json_encode($data);

        // Create the headers for the POST request
        $headers = [
            'Authorization' => 'Bearer ' . $bearerToken,
            'Content-Type'  => 'application/json'
        ];

        // Make the POST request using wp_remote_post
        $response = wp_remote_post($url, [
            'body'    => $dataString,
            'headers' => $headers
        ]);

        // Check for errors in the response
        if (is_wp_error($response)) {
            // Print error message for debugging
            error_log("Error when Subscribe to Google Drive: " . $response->get_error_message());
            return $response;
        }

        // Return the response from the Google Drive API
        return json_decode( wp_remote_retrieve_body($response), true );
    }
    /**
     * Get the start page token for Google Drive changes.
     *
     * @param string $bearerToken The OAuth 2.0 Bearer Token.
     * @return string|null The start page token or null on failure.
     */
    public static function getStartPageToken($bearerToken) {
        // Google Drive API endpoint for getting the start page token
        $url = sprintf( 'https://www.googleapis.com/drive/%s/changes/startPageToken', self::$api_ver );

        // Create the headers for the GET request
        $headers = [
            'Authorization' => 'Bearer ' . $bearerToken,
            'Content-Type'  => 'application/json'
        ];

        // Make the GET request using wp_remote_get
        $response = wp_remote_get($url, [
            'headers' => $headers
        ]);

        // Check for errors in the response
        if (is_wp_error($response)) {
            error_log( 'Error while getting Start Page Token: ' . $response->get_error_message() );
            return null;
        }
        // error_log( 'From get start page token: ' . print_r($response, true) );
        // Decode the response body
        $body = wp_remote_retrieve_body($response);
        $data = json_decode($body, true);

        // Return the start page token
        return isset($data['startPageToken']) ? $data['startPageToken'] : null;
    }

    /**
     * Get a list of changes from Google Drive.
     *
     * @param string $bearerToken The OAuth 2.0 Bearer Token.
     * @param string $pageToken The token to specify the starting point for listing changes.
     * @param int $pageSize The maximum number of changes to return per page.
     * @return mixed The response from the Google Drive API or WP_Error on failure.
     */
    public static function getChanges($bearerToken, $pageToken, $pageSize = 100) {
        // Google Drive API endpoint for listing changes
        $url = sprintf( 'https://www.googleapis.com/drive/%s/changes', self::$api_ver );

        // Create the full URL with query parameters using add_query_arg
        $fullUrl = add_query_arg([
            'pageToken' => $pageToken,
            'pageSize'  => $pageSize
        ], $url);

        // Create the headers for the GET request
        $headers = [
            'Authorization' => 'Bearer ' . $bearerToken,
            'Content-Type'  => 'application/json'
        ];

        // Make the GET request using wp_remote_get
        $response = wp_remote_get($fullUrl, [
            'headers' => $headers
        ]);

        // Check for errors in the response
        if (is_wp_error($response)) {
            error_log('Error when getChanges: ' . $response->get_error_message());
            return $response;
        }

        // Return the response from the Google Drive API
        $body = json_decode( wp_remote_retrieve_body($response), true );

        return $body;
    }

    /**
     * Generate a UUID (version 4).
     *
     * @return string The generated UUID.
     */
    private static function generateUUID() {
        return sprintf(
            '%04x%04x-%04x-%04x-%04x-%04x%04x%04x',
            mt_rand(0, 0xffff), mt_rand(0, 0xffff),
            mt_rand(0, 0xffff),
            mt_rand(0, 0x0fff) | 0x4000,
            mt_rand(0, 0x3fff) | 0x8000,
            mt_rand(0, 0xffff), mt_rand(0, 0xffff), mt_rand(0, 0xffff)
        );
    } 
    private static function appendMoreTokenFields( $token )
    {
        $token['created_at'] = time();
        return $token;
    }
}
